from district42.errors import make_already_declared_error, make_invalid_type_error
from revolt.errors import make_substitution_error

__all__ = ("make_invalid_type_error", "make_already_declared_error", "make_substitution_error",)



from district42 import Props, PropsType, register_type
from district42.types import Schema
from th import PathHolder
from valera import Formatter, ValidationResult

from ._custom_type import CustomSchema

__all__ = ("CustomSchema", "Schema", "Props", "PropsType", "register_type",
           "ValidationResult", "PathHolder", "Formatter",)



def test_district42():
    from d42 import from_native, make_required, optional, register_type, schema  # noqa


def test_blahblah():
    from d42 import fake  # noqa


def test_valera():
    from d42 import ValidationException, validate, validate_or_fail  # noqa


def test_revolt():
    from d42 import schema
    assert schema.str % "revolt" == schema.str("revolt")


def test_custom_type():
    from d42.custom_type import (  # noqa
        CustomSchema,
        Formatter,
        PathHolder,
        Props,
        PropsType,
        Schema,
        ValidationResult,
        register_type,
    )


def test_custom_type_visitors():
    from d42.custom_type.visitors import Generator, Representor, Substitutor, Validator  # noqa


def test_custom_type_errors():
    from d42.custom_type.errors import (  # noqa
        DeclarationError,
        SubstitutionError,
        TypeValidationError,
        ValidationError,
        ValueValidationError,
    )


def test_custom_type_utils():
    from d42.custom_type.utils import (  # noqa
        make_already_declared_error,
        make_invalid_type_error,
        make_substitution_error,
    )



from blahblah import Generator
from district42.representor import Representor
from revolt import Substitutor
from valera import Validator

__all__ = ("Representor", "Generator", "Validator", "Substitutor",)



from unittest.mock import Mock, call, sentinel

from baby_steps import given, then, when
from blahblah import Random, RegexGenerator
from pytest import raises

from d42.custom_type import CustomSchema, Props
from d42.custom_type.visitors import Generator, Representor, Substitutor, Validator


def test_custom_type_abc():
    with when, raises(Exception) as exception:
        CustomSchema()

    with then:
        assert exception.type is TypeError
        assert str(exception.value) == "Cannot instantiate abstract class CustomSchema"


def test_custom_type():
    with given:
        class CustomType(CustomSchema[Props]):
            pass

    with when:
        custom_type = CustomType()

    with then:
        assert isinstance(custom_type, CustomSchema)


def test_custom_type_represent():
    with given:
        mock = Mock(return_value=sentinel.visited)

        class CustomType(CustomSchema[Props]):
            __represent__ = mock

        representor = Representor()
        custom_type = CustomType()

    with when:
        res = representor.visit(custom_type)

    with then:
        assert res is sentinel.visited
        assert mock.mock_calls == [
            call(representor, indent=0)
        ]


def test_custom_type_represent_default():
    with given:
        class CustomType(CustomSchema[Props]):
            pass

        representor = Representor()
        custom_type = CustomType()

    with when:
        res = representor.visit(custom_type)

    with then:
        assert res == "<CustomType>"


def test_custom_type_generate():
    with given:
        mock = Mock(return_value=sentinel.visited)

        class CustomType(CustomSchema[Props]):
            __generate__ = mock

        random = Random()
        generator = Generator(random, RegexGenerator(random))
        custom_type = CustomType()

    with when:
        res = generator.visit(custom_type)

    with then:
        assert res is sentinel.visited
        assert mock.mock_calls == [
            call(generator)
        ]


def test_custom_type_generate_error():
    with given:
        class CustomType(CustomSchema[Props]):
            pass

        random = Random()
        generator = Generator(random, RegexGenerator(random))
        custom_type = CustomType()

    with when, raises(Exception) as exception:
        generator.visit(custom_type)

    with then:
        assert exception.type is NotImplementedError
        assert str(exception.value) == "CustomType has no attribute '__generate__'"


def test_custom_type_validate():
    with given:
        mock = Mock(return_value=sentinel.visited)

        class CustomType(CustomSchema[Props]):
            __validate__ = mock

        validator = Validator()
        custom_type = CustomType()

    with when:
        res = validator.visit(custom_type, value=sentinel.value, path=sentinel.path)

    with then:
        assert res is sentinel.visited
        assert mock.mock_calls == [
            call(validator, value=sentinel.value, path=sentinel.path)
        ]


def test_custom_type_validate_error():
    with given:
        class CustomType(CustomSchema[Props]):
            pass

        validator = Validator()
        custom_type = CustomType()

    with when, raises(Exception) as exception:
        validator.visit(custom_type, value=sentinel.value, path=sentinel.path)

    with then:
        assert exception.type is NotImplementedError
        assert str(exception.value) == "CustomType has no attribute '__validate__'"


def test_custom_type_substitute():
    with given:
        mock = Mock(return_value=sentinel.visited)

        class CustomType(CustomSchema[Props]):
            __substitute__ = mock

        substitutor = Substitutor()
        custom_type = CustomType()

    with when:
        res = substitutor.visit(custom_type, value=sentinel.value)

    with then:
        assert res is sentinel.visited
        assert mock.mock_calls == [
            call(substitutor, value=sentinel.value)
        ]


def test_custom_type_substitute_error():
    with given:
        class CustomType(CustomSchema[Props]):
            pass

        substitutor = Substitutor()
        custom_type = CustomType()

    with when, raises(Exception) as exception:
        substitutor.visit(custom_type, value=sentinel.value)

    with then:
        assert exception.type is NotImplementedError
        assert str(exception.value) == "CustomType has no attribute '__substitute__'"



from district42.errors import DeclarationError
from revolt.errors import SubstitutionError
from valera.errors import TypeValidationError, ValidationError, ValueValidationError

__all__ = ("DeclarationError", "SubstitutionError", "ValidationError",
           "TypeValidationError", "ValueValidationError",)



import sys
from typing import Any, cast, final

if sys.version_info < (3, 11):
    Self = Any
else:
    from typing import Self

from niltype import Nil, Nilable

from d42.custom_type import PathHolder, PropsType, Schema, ValidationResult
from d42.custom_type.visitors import Generator, Representor, Substitutor, Validator

__all__ = ("CustomSchema",)


class CustomSchema(Schema[PropsType]):
    def __init__(self, props: Nilable[PropsType] = Nil) -> None:
        if type(self) is CustomSchema:
            raise TypeError(f"Cannot instantiate abstract class {self.__class__.__name__}")
        super().__init__(props)

    @final
    def __district42__(self, visitor: Representor, *, indent: int = 0, **kwargs: Any) -> str:
        if represent_method := getattr(self, "__represent__", None):
            return cast(str, represent_method(visitor, indent=indent, **kwargs))
        return f"<{self.__class__.__name__}>"

    @final
    def __blahblah__(self, visitor: Generator, **kwargs: Any) -> Any:
        if generate_method := getattr(self, "__generate__", None):
            return generate_method(visitor, **kwargs)
        raise NotImplementedError(f"{self.__class__.__name__} has no attribute '__generate__'")

    @final
    def __valera__(self, visitor: Validator, *, value: Any = Nil, path: Nilable[PathHolder] = Nil,
                   **kwargs: Any) -> ValidationResult:
        if validate_method := getattr(self, "__validate__", None):
            res = validate_method(visitor, value=value, path=path or visitor.make_path(), **kwargs)
            return cast(ValidationResult, res)
        raise NotImplementedError(f"{self.__class__.__name__} has no attribute '__validate__'")

    @final
    def __revolt__(self, visitor: Substitutor, *, value: Any = Nil, **kwargs: Any) -> Self:
        if substitute_method := getattr(self, "__substitute__", None):
            return cast(Self, substitute_method(visitor, value=value, **kwargs))
        raise NotImplementedError(f"{self.__class__.__name__} has no attribute '__substitute__'")



import blahblah
import district42
import revolt  # noqa: F401
import valera

try:
    make_required = district42.make_required
except AttributeError:
    from typing import Any

    def make_required(schema: Any, keys: Any = None) -> Any:
        raise ValueError("make_required is not available in district42 < 1.7.1")

schema = district42.schema
optional = district42.optional
from_native = district42.from_native
register_type = district42.register_type

fake = blahblah.fake

validate = valera.validate
validate_or_fail = valera.validate_or_fail
ValidationException = valera.ValidationException

substitute = revolt.substitute

__all__ = (
    "schema", "optional", "from_native", "register_type", "make_required",
    "fake",
    "validate", "validate_or_fail", "ValidationException",
    "substitute",
)
__version__ = "1.7.1"
